package Chandra::Tools::dmcoords;

use strict;
use warnings;

use vars qw( $VERSION );

$VERSION = '0.21';

use Carp;
use Expect::Simple;

use Env::Path;

# Preloaded methods go here.

sub new
{
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $obj  = {
	      Timeout => 1000,
	      Debug => 0,
	      Verbose => 0,
	      File => undef,
	      Asolfile => undef
	     };

  bless ($obj, $class);

  $obj->{File} = shift;

  croak( __PACKAGE__, ": must specify File attribute\n" )
    unless defined $obj->{File};

  my $attr = pop @_ if @_ && 'HASH' eq ref $_[-1];

  $obj->{Asolfile} = shift;

  if ( defined $attr )
  {
    while( my ( $attr, $val ) = each %{$attr} )
    {
      croak( __PACKAGE__, ": attribute error : `$attr' is not recognized \n" )
	unless exists $obj->{$attr};

      $obj->{$attr} = $val;
    }
  }

  eval { $obj->_connect; };

  croak (__PACKAGE__, ': ', $@) if $@;
  croak (__PACKAGE__, ': Unexpected cruft before the prompt...',
	 $obj->{conn}->before ) if $obj->{conn}->before;

  $obj->deg;
  $obj->status;

  return $obj;
}


sub coords
{
  my $obj = shift;

  unless ( ref( $obj ) )
  {
    my $file = shift;
    # this is gross, but we can escape backwards incompatibility this way
    my $asolfile = ! ref $_[0] && $_[0] !~ /^(sky|chip|cel|det|tdet)$/i ? shift : undef;
    $obj = Chandra::Tools::dmcoords->new($file, $asolfile);
  }

  my @data;
  eval { @data = $obj->_process_coords(@_) };
  croak (__PACKAGE__, ': ', $@) if $@;

  @data;
}

sub _process_coords
{
  my $obj = shift;
  my @coords = @_;

  my @data;
  my $i;
  while ( @coords )
  {
    my $coord = shift @coords;
    if ( ref( $coord ) )
    {
      croak ( "process_coords - illegal argument in coordinate list at index $i\n" )
	unless 'ARRAY' eq ref ($coord);
      push @data, $obj->_process_coord( @{$coord} );
      $i++;
    }

    else
    {
      push @data, $obj->_process_coord( $coord, splice( @coords, 0, 2) );
      $i += 2;
    }
  }

  @data;
}

our %set_values =
  (
   spacecraft => 1,
   telescope => 1,
   detector => 1,
   nominal => 2,
   aspect => 3,
   sim => 2,
   grating => 1,
   'tel/hpy' => 2,
   chip => 1,
   fam => 6,
   'tel/align' => 3,
   'sim/align' => 3,
   'sim/displace' => 6
  );

# this is horrible, but we need to restrict set options to the above
# list so that we can check the number of arguments. otherwise,
# we'll hang if we send fewer than necessary as dmcoords will query
# for the missing ones.  we could query dmcoords for the options it
# supports, but i'm too lazy to do that right now.

sub set
{
  my $obj = shift;

  my $val = shift;

  croak( __PACKAGE__, "->set: unknown value: $val" )
    unless exists $set_values{ lc $val };

  $val = lc $val;

  croak( __PACKAGE__, "->set: expected $set_values{$val} arguments, got ",
       scalar @_)
    unless @_ == $set_values{ lc $val };

  eval {   $obj->{conn}->send( join("\n", "set $val", @_ ) ); };
  croak (__PACKAGE__, ': ', $@) if $@;
  (my $res = $obj->{conn}->before) =~ tr/\r//d;

  my @lines = split( "\n", $res );

  # will see echo of command + 1 line per value sent + 1 line of all of the prompts.
  # anything else is bogus
  splice( @lines, 0, 2 + $set_values{$val} );

  croak (__PACKAGE__, '->set: ', shift @lines) if @lines;
}

sub status
{
  eval {   $_[0]->{conn}->send( "status" ); };
  croak (__PACKAGE__, ': ', $@) if $@;
}

sub deg
{
  eval {   $_[0]->{conn}->send( "deg" ); };
  croak (__PACKAGE__, ': ', $@) if $@;
}

sub _process_coord
{
  my $obj = shift;
  my @coord = @_;

  $obj->{conn}->send( join( ' ', @coord ) );

  my $res = $obj->{conn}->before;
  $res =~ tr/\r//d;

  my @lines = split( "\n", $res );
  # first line is just the command just sent (due to echoing)
  shift @lines;

  my %data;
  eval
  {
    foreach ( @lines )
    {
      if ( /^\s*\(\s*RA/i )
      {
	if ( /\d:\d/ )
	{
	  die( "$_\n" )
	    unless 
	      @{$data{cel}{hms}}{ qw( ra dec )} = 
		/\s+(\d[\d.:]+)[^\d+-]+([+-]?\d[\d.:]+)/;
	}
	else
	{
	  die( "$_\n" )
	    unless @{$data{cel}{deg}}{ qw( ra dec )} = /\s+(\d[\d.]+)\s+([+-]?\d[\d.]+)/;
	}
      }

      elsif ( /^\s*zo/i )
      {
	if ( /sky/i )
	{
	  die( "$_\n" )
	    unless @{$data{z0}{sky}}{qw{ x y}} =
	      /\s+([+-]?\d[\d.]+)\s+([+-]?\d[\d.]+)/;
	}
	
	elsif ( /det/i )
	{
	  die( "$_\n" )
	    unless @{$data{z0}{det}}{qw{ x y}} = 
	      /\s+([+-]?\d[\d.]+)\s+([+-]?\d[\d.]+)/;
	}
	
	elsif ( /ra/i )
	{
	  if ( /\d:\d/ )
	  {
	    die( "$_\n" )
	      unless 
		@{$data{z0}{hms}}{ qw( ra dec )} = 
		  /\s+([+-]?\d[\d.:]+)[^\d+-]+([+-]?\d[\d.:]+)/;
	  }
	  else
	  {
	    die( "$_\n" )
	      unless @{$data{z0}{deg}}{ qw( ra dec )} =
		/\s+([+-]?\d[\d.]+)\D+([+-]?\d[\d.]+)/;
	  }
	}
	
      }

      elsif ( /theta/i )
      {
	my ( $theta, $theta_unit, $phi, $phi_unit );
	die( "$_\n" )
	  unless
	    ( $theta, $theta_unit, $phi, $phi_unit ) =
	      /\s+(\d[\d.]+)\s*(\S+)\s+(\d[\d.]+)\s*(\S+)/;

	# convert everything to degrees
	eval { 
	  $data{theta} = _angle_cvt( $theta, $theta_unit );
	  $data{phi}   = _angle_cvt( $phi, $phi_unit );
	};

	if ( $@ ) { chomp $@; die "$@: $_\n" }
      }

      elsif ( /energy/i )
      {
	die( "$_\n" )
	  unless
	    ( $data{energy} ) = /(\d[\d.]+)/;
      }

      else
      {
	my ( $what, $x, $y );
	
	die( "$_\n" )
	  unless 
	    ( $what, $x, $y ) = /(.*)\s+([+-]?\d[\d.]+)\s+([+-]?\d[\d.]+)/;
	
	if ( $what =~ /sky/i )
	{
	  $data{sky} = { x => $x, 'y' => $y };
	}
	
	elsif ( $what =~ /detx/i )
	{
	  $data{det} = { x => $x, 'y' => $y };
	}
	
	elsif ( $what =~ /chip/i )
	{
	  my (undef, $id) = split ' ', $what;
	  $data{chip} = { x => $x, 'y' => $y, id => $id };
        }
	
	elsif ( $what =~ /tdet/i )
	{
	  $data{tdet} = { x => $x, 'y'=> $y };
	}
	
	elsif ( $what =~ /logical/i )
	{
	  $data{logical} = { x => $x, 'y'=> $y };
	}
	
	elsif ( $what =~ /gdp/i )
	{
	  $data{gdp} = { x => $x, 'y'=> $y };
	}
	
	elsif ( $what =~ /gac/i )
	{
	  $data{gac} = { r => $x, d=> $y };
	}
	
	else
	{
	  die( "$_\n" )
	}
	
      }
    }
  };

  die( 'parse error for (', join( ',', map { "'$_'" } @coord), '): ', $@ )
    if $@;

  return \%data;
}

# convert into degrees
sub _angle_cvt
{
  my ( $value, $unit ) = @_;

  # seconds of arc
  if ( $unit eq '"' )
  {
    $value /= 3600;
  }

  elsif ( $unit eq "'" )
  {
    $value /= 60;
  }

  elsif ( $unit eq 'deg' )
  {
  }

  else
  {
    die( "unknown unit: $unit\n" );
  }

  $value;
}

sub _connect
{
  my $obj = shift;

  # let's find that dmcoords program!
  croak( "unable to find dmcoords program in PATH! Please fix!\n" )
    unless Env::Path->PATH->Whence( 'dmcoords' );

  my $cmd = 
    join( ' ',
	  'dmcoords',
	  'verbose=1',
	  'option=',
	  "infile=$obj->{File}",
	  'asolfile=' . (defined($obj->{Asolfile}) ? $obj->{Asolfile} : 'none')
	);

  eval {
    $obj->{conn} = new Expect::Simple
    { Cmd => $cmd,
      Prompt => [ -re => 'dmcoords>:\s+' ],
      DisconnectCmd => 'q',
      Verbose => $obj->{Verbose},
      Debug => $obj->{Debug},
      Timeout => $obj->{Timeout}
    }
  };

  if ( $@ )
  {
    my $err;
    ( $err = $@) =~ s/Expect::Simple//;
    croak( "connect error: $err" );
  }

}

sub _disconnect
{
  delete shift()->{conn};
}

sub DESTROY
{
  shift()->_disconnect;
}


1;

__END__


=head1 NAME

Chandra::Tools::dmcoords - interface to CIAO dmcoords command

=head1 SYNOPSIS

  use Chandra::Tools::dmcoords;

  @req = ( [ sky => ( 10, 10 )  ],
	      [ chip => ( 10, 10 ) ],
	      [ cel => ( '02:19:40.827', '+29:47:17.80' ) ],
	      [ det => ( 10, 10 ) ] );

  @results = Chandra::Tools::dmcoords->coords( $file, @req );

  $dmcoord = new Chandra::Tools::dmcoords( $file [, $asolfile] [, \%attr] );
  @data = $dmcoord->coords( 'sky' => (10, 10 ) );
  @data = $dmcoord->coords( 'chip' => (10, 10 ) );
  $dmcoord->set( chip => 'acis-s3' );

=head1 DESCRIPTION

B<Chandra::Tools::dmcoords> provides a wrapper around the CIAO
B<dmcoords> command, enabling it to be used as a subroutine.  There
are two modes of operation.  In the first, an instance of B<dmcoords>
is started up, a list of coordinates is fed to it, and the instance of
B<dmcoords> is shut down.  In the second mode, more of a server mode,
B<dmcoords> is started up, and multiple, separate, coordinate requests
may be sent.  The instance of B<dmcoords> will be shut down when the
handle to the Chandra::Tools::dmcoords object goes out of scope (or is
explicitly undefined).

Exceptions will be thrown on error (match with C</^Chandra::Tools::dmcoords/>).

=head2 Working around dmcoords bugs

Workarounds to several bugs in B<dmcoords'> interactive mode (as of
CIAO 3.2.1) are implemented.  When B<dmcoords> is first started, the
C<deg> and C<status> commands are sent to it.  This magically clears
up problems with data entry and HRC pixel size issues.

=head2 Methods

=over 8

=item new

    $obj = Chandra::Tools::dmcoords->new($file);
    $obj = Chandra::Tools::dmcoords->new($file, \%attr);
    $obj = Chandra::Tools::dmcoords->new($file, $asolfile );
    $obj = Chandra::Tools::dmcoords->new($file, $asolfile, \%attr );

Create a new B<Chandra::Tools::dmcoords> object, starting up an
instance of B<dmcoords> and leaving it running.  Multiple copies of
B<dmcoords> may be started by creating new objects.  At a minimum the
name of the events file should be passed.  If B<dmcoords> should
correct for FID light drifits, then the aspect solution file should be
passed as well.  This may be a stack (in which case
the filename which have the standard C<@> prefix).  The following
additional attributes may be specified via the C<%attr> hash:

=over

=item Timeout

See the B<Expect::Simple> docs.  Defaults to 1000.

=item Debug

See the B<Expect::Simple> docs.  Defaults to 0.

=item Verbose

See the B<Expect::Simple> docs.  Defaults to 0.

=back

=item coords

    # as object method
    @data = $obj->coords( @coords );

    # as class method
    @data = Chandra::Tools::dmcoords->coords( $file, @coords );
    @data = Chandra::Tools::dmcoords->coords( $file, $asolfile, @coords );

When called as an object method, it passes the coordinate list to the
running B<dmcoords> associated with that object.  When called as
a class method, it starts up B<dmcoords>, passes it the coordinate list,
and then shuts it down.

The coordinate list contains one or more coordinates to be transformed
by dmcoords.  A coordinate is either a list of three elements or a
reference to a list of three elements.  The elements are the input
coordinate system, and the two input coordinates. For instance,

  @coords = ( 'sky', 10, 10, 'chip', 20, 20  );
  @coords = ( sky => (10, 10), chip => ( 20, 20 ) );
  @coords = ( [ sky => ( 10, 10 ) ], [ chip => ( 20, 20 ) ] );

(The first two forms are actually syntactically equivalent to Perl;
the second is just more obvious).  The array reference version makes
it easier to index into the B<coords> array, if the calling program needs to.
One can mix the list and array reference modes,

  @coords = ( sky => (10, 10), [ chip => ( 20, 20 ) ] );

if one is masochistic.

The available input coordinate systems at the time of this writing are
C<sky>, C<cel>, C<chip>, C<det>.  C<tdet> doesn't seem to be available
in B<dmcoords> interactive mode (which is used by this module) for
some reason.  Note that this method doesn't do any error checking; it
just passes things along to B<dmcoords>.  If it can't parse
B<dmcoords> output, it'll B<croak>.

B<coords> returns a list of hash references, one per input object.
Each hash contains all of the information B<dmcoords> is capable of
producing.  For example,

   use Data::Dumper;
   print Dumper [ $dmcoord->coords( 'sky', 20, 20  ) ];

may produce

  $VAR1 = [
         {
            'sky' => {
                       'x' => '20.00',
                       'y' => '20.00'
                     },
            'logical' => {
                           'x' => '20.00',
                           'y' => '20.00'
                         },
            'chip' => {
                        'id' => 'ACIS-S4',
                        'x' => '-5941.18',
                        'y' => '6419.76',
                      },
            'det' => {
                       'x' => '20.00',
                       'y' => '20.00'
                     },
            'phi' => '225.00',
            'tdet' => {
                        'x' => '0.00',
                        'y' => '0.00'
                      },
            'cel' => {
                       'deg' => {
                                  'dec' => '18.24039',
                                  'ra' => '19.24235'
                                },
                       'hms' => {
                                  'dec' => '+18:14:25.40',
                                  'ra' => '01:16:58.164'
                                }
                     },
            'theta' => '47.270'
          }
        ];


The C<theta> and C<phi> results are converted into degrees, for your
convenience.


=item set

  $dmcoord->set( option => value );

This is available as an object method only.  This sets one of B<dmcoords>
internal configuration values, which will affect subsequent calls to
the B<coords> method.  The following options were available at the time
of writing (only these will be recognized):

C<spacecraft>,
C<telescope>,
C<detector>,
C<nominal>,
C<aspect>,
C<sim>,
C<grating>,
C<tel/hpy>,
C<chip>,
C<fam>,
C<tel/align>,
C<sim/align>,
C<sim/displace>.

Note that the value for C<chip> has the format C<det-chipid> where
C<det> is C<hrc> or C<acis>. C<chipid> is either 

=over 

=item * 

a number: 0-9 for ACIS, 0-3 for HRC.

=item *

a subdetector( C<i> or C<s>) and a number 

  ACIS-I: 0-3
  ACIS-S: 0-5 
  HRC-I: 0
  HRC-S: 1-3

=back

For example: 

	acis-s3
	acis-9
	hrc-0
	hrc-s1

=back


=head1 PREREQUISITES

B<Expect::Simple>, B<Expect>, and B<dmcoords>, of course.

=head1 BUGS

The other interactive commands to B<dmcoords> (C<set>) are not
supported.  You can send them, but the module won't be prepared to
parse stuff and will (possibly) go belly up and start WWIII.

=head1 AUTHOR

Diab Jerius ( djerius@cfa.harvard.edu )


=cut
