package Data::ParseTable;

use 5.008;
use strict;
use warnings;
#use diagnostics;
use Carp;
use IO::All;
use Text::RecordParser;
use File::MMagic;
use HTML::TableExtract;
use PDL::Lite;
use IO::All;
use subs qw(dbg);

use POSIX qw(strtod);

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw();
our @EXPORT_OK = qw( parse_table );
%EXPORT_TAGS = ( all => \@EXPORT_OK );

our $VERSION = '0.06';

our $Debug = 0;

# Preloaded methods go here.


sub parse_table{

    my $data_in = shift;

    my $opt_in = shift;

    my $parser = Data::ParseTable->new($data_in, $opt_in);

    my $data_out_ref = $parser->get_data_out();

    return $data_out_ref;

}


sub new{

    my $classname = shift;
    my $self = {};
    bless $self, $classname;

    my $data_in = shift;
    my $opt_ref = shift;

    $self->{data_in} = $data_in;

    # get and set options
    $self->set_options($opt_ref);

    # let's store them more handily    
    my $opt = $self->{options};

    if ($opt->{debug}){
	$Debug = 1;
    }

   # Now let's check the to see if we've passed a filename or an array reference for 
   # the first argument

    my $input_type = ref($data_in);
    my $bad_data_input = __PACKAGE__ . ": parse_data not filename or array reference \n";


    $opt->{passed_filename} = ($input_type =~ /ARRAY.*/) ? 0
	                    : ($input_type =~ /\s*/ )    ? 1	
	                    :                            croak($bad_data_input)
                            ;

    

   # If passed a filename, let's confirm that the file actually exists, is a file, and  
   # is readable
    
    if ($opt->{passed_filename}){
	$self->parse_file();
    }
    else{
	$self->parse_lines();
    }

    # fits are handled by their own routine
    unless ($self->{options}->{format} eq 'fits'){
	if ($self->{options}->{return_as_pdl}){
	    $self->convert_to_pdl($self->{data_out});
	}
    }
    
    return $self;

}

sub parse_file{

    my $self = shift;
    my $filename = $self->{data_in};
    my $opt = $self->{options};

    # check to make sure file is readable
    croak(__PACKAGE__ . ": file $filename is not a readable file \n") 
	unless ((-r $filename) and ((-e $filename) or (-l $filename)));

    $opt->{is_binary} = $self->is_binary();

    if (defined $opt->{format}){
	$self->parse_defined_file();
    }
    else{
	$self->parse_guess_file();
    }

}

sub parse_lines{

    my $self = shift;
    my $opt = $self->{options};
    
    $opt->{is_binary} = $self->is_binary();
    
    if ($opt->{is_binary}){
	croak(__PACKAGE__ . ": Binary file may not be passed as array\n");
    }

    if (defined $opt->{format}){
	$self->parse_defined_lines();
    }
    else{
	$self->parse_guess_lines();
    }

    
}

sub parse_defined_file{
    
    my $self = shift;
    my $opt = $self->{options};

    $self->{data_out} = ($opt->{format} eq 'csv')  ? $self->parse_csv_file()
		      : ($opt->{format} eq 'html') ? $self->parse_html_file()
		      : ($opt->{format} eq 'fits') ? $self->parse_fits_file()
	              : ($opt->{format} eq 'rdb')  ? $self->parse_rdb_file()
   		      :                             croak( __PACKAGE__ . ": specified format $opt->format is not one of (csv|html|fits|rdb) \n")
		      ;
    
} 

sub parse_defined_lines{
    my $self = shift;
    my $opt = $self->{options};
    $self->{data_out} = ($opt->{format} eq 'csv')  ? $self->parse_csv_lines()
	              : ($opt->{format} eq 'html') ? $self->parse_html_lines()
	              : ($opt->{format} eq 'rdb')  ? $self->parse_rdb_lines()
	              :                              croak( __PACKAGE__ . ": specified format for line of $opt->format is not one of (csv|html|rdb) \n")
                      ;
}


                

sub parse_guess_file{

    my $self = shift;
    my $opt = $self->{options};

    my $data_ref;
    eval {
	$data_ref = $self->read_fits_table();
    };
    unless ($@){
	$self->{data_out} = $data_ref;
	$opt->{format} = 'fits';
	return 1;
    }

    my $filename = $self->{data_in};
    
    if ($opt->{is_binary} and not defined $opt->{format}){
	croak( __PACKAGE__ . " file $filename is binary file and not a fits file \n");
    }

    # if not a fits file, lets make it into an array of lines and send it to the other routine


    my @line_array = io($filename)->slurp;
    $self->{data_in} = \@line_array;
    $self->parse_guess_lines;

    return 1;

}

sub parse_csv_file{

    my $self = shift;
    my @data_lines = io($self->{data_in})->slurp();
    $self->{data_in} = \@data_lines;
    $self->parse_csv_lines;
    return $self->{data_out};
}

sub parse_html_file{

    my $self = shift;
    my @data_lines = io($self->{data_in})->slurp();
    $self->{data_in} = \@data_lines;
    $self->parse_html_lines;
    return $self->{data_out};

}
    

sub parse_rdb_file{

    my $self = shift;
    my @data_lines = io($self->{data_in})->slurp();
    $self->{data_in} = \@data_lines;
    $self->parse_rdb_lines;
    return $self->{data_out};    

}

sub parse_rdb_lines{
    my $self = shift;
    $self->{options}->{field_separator} = "\t";
    check_rdb( $self->{data_in}, $self->{options});
    $self->{data_out} = $self->parse_csv_lines;
    return 1;
}



sub parse_fits_file{
    my $self = shift;
    return  $self->read_fits_table();
}



sub parse_guess_lines{

    my $self = shift;

    my $opt_ref = $self->{options};
    my @data_lines = @{$self->{data_in}};

    if(is_fits_start($data_lines[0])){
	croak( __PACKAGE__ . ": FITS files cannot be passed as arrays\n");
    }
    
    if (is_html(\@data_lines)){
	$opt_ref->{format} = 'html';
	$self->{data_out} = $self->parse_html_lines();
	return 1;
    }
    
    my $data_out;

    if ( check_rdb( \@data_lines, $opt_ref )){
	$self->parse_rdb_lines();
	$data_out = $self->{data_out};
    }

    if (not defined $self->{options}->{format}){

#	eval {
	    $data_out = $self->parse_csv_lines();
#	};
#	unless ($@){
	    if (not defined $self->{options}->{format}){
		$self->{options}->{format} = 'csv';
	    }
#	}
    }

    if (not defined $self->{options}->{format}){
	croak(__PACKAGE__ . ": Could not guess file type \n");
    }
	
    $self->{data_out} = $data_out;
    return 1;

}


sub parse_csv_lines{

    my $self = shift;

# I should address the fact that this doesn't work if passed an array without
# carriage returns
    my $text_recordparser;

    my $data_ref = $self->{data_in};
    my $opt_ref = $self->{options};
    
    $data_ref = strip_data_comments($data_ref, $opt_ref);

    my $fs = $opt_ref->{field_separator};

    my $n_fields;

    if (not defined $fs){
	$fs = guess_field_separator($data_ref);
    	$n_fields = count_fields_using_fs($data_ref, $fs);
    }
    else{
	if ("$fs" eq '|'){
	    # override for correct splitting
	    $fs = qr/\|/;
	}

	$n_fields = count_fields_using_fs($data_ref, $fs);
       
	croak(__PACKAGE__ . ": Specified Field Separator qq($fs) does not separate file cleanly") 
	    unless ($n_fields);
    }

    # If whitespace separated, then clean leading/trailing whitespace
    if ($fs eq qr/\s+/) {
	s/^ *| *$//g for @{$data_ref};
    }

    dbg "field sep = :$fs: \n";
    $opt_ref->{field_separator} = $fs;

    my $parsed_data;

    eval{
	my $parser = Text::RecordParser->new;
	$parser->data( $data_ref );
	$parser->field_separator( $fs );
	$parser->bind_header;
	$parsed_data = $parser->fetchall_arrayref( { Columns => {} } );
 	croak(__PACKAGE__ .": Text::RecordParser returned wrong-sized array\n") 
	    unless ( (keys %{$parsed_data->[0]}) == $n_fields );
    };
    if ($@){
        $parsed_data = parse_text_data( $data_ref, $fs);
    }

    croak(__PACKAGE__ . "No text columns found\n") if ((keys %{$parsed_data->[0]}) < 1);
    
    $parsed_data = tidy_ascii_spaces($parsed_data);

    if ($opt_ref->{return_as_hash}){
	my $data_hash_ref = table_array_to_hash($parsed_data);
	return $data_hash_ref;
    }

    return $parsed_data;
}


sub parse_text_data{

    my $data_ref = shift;
    my $fs = shift;

    my @data_lines = @{$data_ref};

    my @result_array;

    chomp $data_lines[0];
    my @header_array = split("$fs", $data_lines[0]);

    for my $line (1 .. $#data_lines){
	my %line_hash;	
	chomp $data_lines[$line];
	my @line_array = split( "$fs", $data_lines[$line]);
	
	for my $field (0 .. $#line_array){
	    $line_hash{$header_array[$field]} = $line_array[$field];

	}
	push @result_array, \%line_hash;

    }

    return \@result_array;
}

sub read_fits_table{

    my $self = shift;
    my $filename = $self->{data_in};
    my $opt_ref = $self->{options};

    eval 'use Astro::FITS::CFITSIO::Simple qw/ rdfits /';
    if ($@){
        croak(__PACKAGE__ .": !$@");
    }

    my %fits_hash = rdfits($filename);

    # Convert from hash with PDLs to a hash of array refs if required
    my $output_hash_ref = ($opt_ref->{return_as_hash} and $opt_ref->{return_as_pdl})
                                   ? \%fits_hash
                                   : convert_pdl_hash(\%fits_hash);

    if ($opt_ref->{return_as_hash}){
        return $output_hash_ref;
    } else {
        my $fits_array_ref = table_hash_to_array($output_hash_ref);
        return $fits_array_ref;
    }

}


sub set_options{
    
    my $self = shift;
    my $opt_ref = shift;

    %{$self->{options}} = (format          => undef,  # File format (supported values are fits,rdb,html or csv)
			   field_separator => undef,  # Field sep. for 'csv' format
			   return_as_hash  => undef,  # Override default to return array of hashes
			   return_as_pdl   => undef,  # Returns PDLs for FITS files when return_as_hash=>1
			   comment         => undef,  # Comment specifier (e.g. qr/^#/)
			   strip_quotes    => 0,
			   debug           => 0,
			   );


    # And populating it with anything passed in

    if (defined $opt_ref){
	while( my ( $option, $value ) = each %{$opt_ref} ){
	    croak( __PACKAGE__ . ": option error : `$option' is not recognized \n" )
		unless exists $self->{options}->{$option};
	    $self->{options}->{$option} = $value;
	}
    }
    
    $Debug = $self->{options}->{debug};
    # Let's check that the options are relatively reasonable
    $self->check_options();

}



sub get_data_out{
    my $self = shift;
    return $self->{data_out};
}


#****************************************************************************************
sub convert_to_pdl {
# Check each column (which must be an array-ref) for non-numeric entries
# by attempting to convert to a double and checking for problems.
# If none are found, convert to pdl
#****************************************************************************************
    my $self = shift;
    my $data = shift;
    local $_;
    
    while (my ($name, $col) = each %{$data}) {
	my $numeric = 1;
	foreach (@{$col}) {
	    $! = 0;
	    my ($num, $n_unparsed) = strtod($_);
	    if ($n_unparsed != 0 || $!) {
		$numeric = 0;
		last;
	    }
	}
	$data->{$name} = pdl $col if $numeric;
    }
}



#*********************************


sub is_binary{

    my $self = shift;
    my $fm = new File::MMagic;
    
    my $type = $self->{passed_filename} ? $fm->checktype_filename($self->{data_in})
	:                                 $fm->checktype_contents($self->{data_in})
	; 

    if ($type eq 'application/octet-stream'){
	return 1;
    }

    return 0;
}


    

#*********************************
sub is_html{
   
    # receive a pointer to a text array and check if it matches general patterns for html
    
    my $data_ref = shift;
    my $whole_file = join(qq{},@$data_ref);    
    
    # HTMl can begin with <!DOCTYPE , <HTML, <!-- comments, or whitespace
    # I'd like to find a module that does this for me... it can also begin with
    # other funny stuff like xml
    if (($whole_file =~ /\s*<!DOCTYPE\s*HTML.*/i) or 
	($whole_file =~ /\s*<HTML>.*/i) or
	($whole_file =~ /(\s*<!--.*-->\s*)*<HTML>/i)){
	return 1;
    }
    return 0;
    
    
}

#*********************************
sub is_fits_start{
    
    my $first80 = shift;

    # return 1 if matches SIMPLE
    my $isfits = ($first80 =~ /^SIMPLE\s+=\s*T.*/) ? 1 : 0 ;
    return $isfits;
}


#*********************************
sub parse_html_lines{

    my $self = shift;

    # for a table of html data, return a reference to a hash of that
    # data

    
    my $data_ref = $self->{data_in};
    my $opt_ref = $self->{options};

    
    my $whole_file = join(qq{},@$data_ref);

    # strip out newlines because of apparent bug in HTML::TableExtract
    $whole_file =~ s/\n//g;
    
    my $te = HTML::TableExtract->new( depth => 0 );
#    my $te = HTML::TableExtract->new();
    $te->parse($whole_file);

    # stolen code from get_iFOT_events.pl

    # Insist that exactly one table match.  (Use more robust error handling eventually)
    my @ts = $te->table_states;
#    print "tables = ", scalar(@ts), "\n";
#    croak("ERROR - more than one HTML table had the required headers") if (@ts > 1);
    croak(__PACKAGE__ . ": Could not find any HTML tables") if (@ts < 1);  # No data in table

    # Find the table coordinates and now parse the entire table, not just the 
    # spec'd header columns
    my @table_coords = $ts[0]->coords;
    my $ts = $te->table_state(@table_coords);

    # Get the data into a useful form (hash of arrays)
    my @rows = $ts->rows();
    my @cols = @{$rows[0]};

    croak(__PACKAGE__ . ": Empty table\n") if (@cols < 1);
    croak(__PACKAGE__ . ": Table contains a null column name\n") if grep {not defined} @cols;

   map { s/\A [\s<>]+ | [\s<>]+ \Z//gx } @cols;  # clean column names 
    my %data;
    foreach my $i (0 .. $#cols) {
        $data{$cols[$i]} = [ map { $_->[$i] } @rows[1..$#rows] ];
    }

    if ($opt_ref->{return_as_hash}){
	$self->{data_out} = \%data;
	return 1;
    }

    my $array_ref = table_hash_to_array(\%data);
    $self->{data_out} = $array_ref;
    return 1;

    
}










#*********************************
sub tidy_ascii_spaces{
 
    my $in_array_ref = shift;
    my @out_array;

    foreach my $hash (@{$in_array_ref}){
	my %newhash;
	while( my ($header, $value) = each %{$hash}){
	    # strip leading or trailing whitespace off fields
	    
	    $header =~ s/^\s+//;
	    $header =~ s/\s+$//;
	    $value =~ s/^\s+//;
	    $value =~ s/\s+$//;

	    # strip external single or double quotes off a header
	    if ($header =~ /^\'(.*)\'$/){
		$header = $1;
	    }
	    elsif ($header =~ /^\"(.*)\"$/){
		$header = $1;
	    }

	    # strip external single or double quotes off a field
	    if ($value =~  /^\'(.*)\'$/){
		$value = $1;
	    }
	    elsif ($header =~ /^\"(.*)\"$/){
		$value = $1;
	    }

	    # make a new hash with the cleaned up text
	    $newhash{$header} = $value;
	}
	push( @out_array, \%newhash);
    }

    # return a reference to the newly created array
    return \@out_array;
}

#*********************************
sub make_data_array{

    my $filename = shift;
    my @data_lines = io($filename)->slurp;
    return @data_lines;

}

#*********************************
sub guess_field_separator{

    my $data_ref = shift;
#    my @field_sep = ("\t", qr/\s+/, "&", "|", ",");
    my @field_sep = (",", "&", qr/\|/, "\t", qr/\s+/);

    my $fallback_fs;

    foreach my $fs (@field_sep){
	my $n_fields = count_fields_using_fs($data_ref, $fs);
	next if ($n_fields == 0);
	if ($n_fields >= 2){
	    return $fs;
	}
	elsif ($n_fields == 1){
	    $fallback_fs = $fs;
	}
    }
    if (defined $fallback_fs){
	return $fallback_fs;
    }
    else {
	croak(__PACKAGE__ . ": Could not determine field separator\n");
    }
}


#*********************************
sub count_fields_using_fs{

    # With "good" defined as "will split the lines of the file into an equal
    # number of columns for all lines"

    my $data_ref = shift;
    my $fs = shift;

    my $length;

    # clean leading/trailing spaces (NOT whitespace) for purpose of counting cols
    my @clean_data_ref = @{$data_ref};
    s/^ *| *$//g for @clean_data_ref;
    
    my $split_nfields;
    my $nfields_A;

    for my $line (@clean_data_ref){
	my @array = split( $fs, $line);
	if (not defined $nfields_A){
	    $nfields_A = scalar(@array);
	}
	$split_nfields = $nfields_A;
	next if (scalar(@array) == $nfields_A);
	$split_nfields = 0;
    }
    my $recparse_nfields;
    my $nfields_B;
    $length = $split_nfields;
   
    eval{
	my $parser = Text::RecordParser->new;    
	$parser->data( \@clean_data_ref );
	$parser->field_separator( $fs );
	my $parsed_data = $parser->fetchall_arrayref;
       	$recparse_nfields = scalar(@{$parsed_data->[0]});

    # not sure what to do with files of single column data
    # might not have a handy way of dealing with that unless
    # explicit...

	foreach my $record (@$parsed_data){
	    if (not defined $nfields_B){
		$nfields_B = scalar(@$record);
	    }
	    $recparse_nfields = $nfields_B;
	    next if ($recparse_nfields == $nfields_B);
	    
	    $recparse_nfields = 0;
	}
    };
    unless ($@){
	if ($recparse_nfields > $split_nfields){
	    $length = $recparse_nfields;
	}
    }

    dbg " Using $fs, we've got $length fields \n";
 
    return $length;
}


#*********************************
sub table_array_to_hash{
    
    my $array_ref = shift;
    my $arr_length =  scalar(@{$array_ref});
    my @columns = keys %{$array_ref->[0]};

    my %hash_conv;

    for my $i (0 .. ($arr_length-1)){
	foreach my $header (@columns){
	    if (not defined ($hash_conv{$header})){
		$hash_conv{$header} = [];
	    }
	    push @{$hash_conv{$header}}, $array_ref->[$i]->{$header};
	}
    }

    return \%hash_conv;
}

#*********************************
sub table_hash_to_array{

    
    my $hash_ref = shift;
    my @columns = keys %{$hash_ref};
    my @array;	
    my $hash_arr_length = scalar(@{$hash_ref->{$columns[0]}});
    
    for my $i (0 .. ($hash_arr_length-1)){
	my %line;
	foreach my $header (@columns){
	    $line{$header} = $hash_ref->{$header}[$i];
	}
	push @array, \%line;
    }

    return \@array;

}

#*********************************
sub table_print_hash{

    my $hash_ref = shift;
    my @columns = keys %{$hash_ref};

    foreach my $header (@columns){
	print "$header: ";
	my $line = join(':,:', @{$hash_ref->{$header}});
	print "$line \n";
    }
}

#*********************************
sub table_print_array{

    my $array_ref = shift;
    my $arr_length = scalar(@{$array_ref});
    
    for my $i (0 .. ($arr_length-1)){
	print "line $i: ";
	while (my ($header, $data) = each %{$array_ref->[$i]}){
	    print "$header = :$data:, ";
	}
	print "\n";
    }	

}

#*********************************
sub check_rdb{

    my $data_ref = shift;
    my $opt_ref = shift;
    
    my $secondline = $data_ref->[1];
    my $is_rdb;

    if ($secondline =~ /^(\d*\w\t)+/){
	$is_rdb = 1;
    }
    if ($secondline =~ /^(-+\t)+/){
	$is_rdb = 1;
    }

    if (defined $opt_ref->{format}){
	if (($opt_ref->{format} eq 'rdb') and (not $is_rdb)){
	    croak(__PACKAGE__ . ":RDB specified, but does not match second line pattern \n");
	}
    }

    if ($is_rdb){
	$opt_ref->{format} = 'rdb';
	splice(@{$data_ref}, 1, 1); 
	return 1;
    }
    else{
	return 0;
    }
    

}

#*********************************
sub strip_data_comments{

    my $data_lines_ref = shift;
    my $opt_ref = shift;

    my $comment = $opt_ref->{comment};
    my @cleandata;
    
    # if a comment symbol has been specified
    # remove lines that start with the symbol
    # and return a reference to a new line array

    if (defined $comment){
	my @cleandata;
	@cleandata = grep(!/^$comment/, @{$data_lines_ref});
	return \@cleandata;
    }
    # if no comment symbol is specified, just
    # return the original array reference
 
    return $data_lines_ref;
}


#*********************************
sub convert_pdl_hash{

    
    my $pdl_hash = shift;
 
    my %not_pdl_hash;

    my $pdl_type;   
    
    eval 'use PDL qw/ list / ' ;
    if ($@){
	croak(__PACKAGE__ . ": !$@");
    }


    while ( my ($header, $value) = (each %{$pdl_hash})){

	if (UNIVERSAL::isa($value,'PDL')) {
	    my $pdl_dim = $value->getndims;
	    if ($pdl_dim <= 1) {
		$not_pdl_hash{$header}  = [ PDL::list($value) ];
	    } else {
		# Caller is gonna get some pdls
		my $nrows = ($value->dims)[-1];
		my $slice = join(',', (':') x ($pdl_dim-1));
		$not_pdl_hash{$header} =
		  [ map { $value->slice($slice . ",$_")->squeeze } (0..$nrows-1) ];
	    }
	}
	else {
	    # Not a PDL, so just return as an array ref
	    $not_pdl_hash{$header} = ref($value) eq 'ARRAY' ? $value : [ $value ];
	}       
    }


    return \%not_pdl_hash;

}

#*********************************
sub check_options{

    my $self = shift;
    my $opt_ref = $self->{options};

    if (defined $opt_ref->{format}){
	croak( __PACKAGE__ . ": Invalid format specified: should be 'html','fits','csv', or 'rdb' \n")
	    unless ($opt_ref->{format} =~ /(html|fits|csv|rdb)/);
    }
    
    if (defined $opt_ref->{comment} and (not defined $opt_ref->{field_separator})){
	croak( __PACKAGE__ . ": Comment symbol cannot be possible field separator \n") 
	    if ($opt_ref->{comment} =~  /(,|&|\||\\t|\s+)/ );
    }
    
    if (defined $opt_ref->{comment} and defined $opt_ref->{field_separator}){
	croak( __PACKAGE__ . ": Comment symbol cannot match field separator \n")
	    if ($opt_ref->{comment} eq $opt_ref->{field_separator});
    }
    

}

##*********************************
sub dbg  {

    if ($Debug) {
	my $args = join('',@_) || "";
	my $caller = (caller(1))[0];
	my $line = (caller(0))[2];
	$caller ||= $0;
	if (length $caller > 22) {
	    $caller = substr($caller,0,10)."..".substr($caller,-10,10);
	}
	my $align = ' 'x39;
	$args =~ s/\n/\n$align/g;
    print STDERR sprintf ("%02d:%02d:%02d [%22.22s %4.4s]  %s\n",
                          (localtime)[2,1,0],$caller,$line,$args);
    }
}

1;


__END__


=head1 NAME

Data::ParseTable - Perl extension for table parsing

=head1 SYNOPSIS

Common Usage

  use Data::ParseTable;

  $return_array_of_hashes = Data::ParseTable::parse_table( $filename );

  $return_hash_array =      Data::ParseTable::parse_table( $filename , 
                                                          { return_as_hash => 1 });

  $return_array_of_hashes = Data::ParseTable::parse_table( \@data_lines );

  $return_hash_array =      Data::ParseTable::parse_table( \@data_lines , 
                                                          { return_as_hash => 1 });



=head1 DESCRIPTION

Data::ParseTable parses text, ascii or binary fits, or html tables into arrays of hashes or hashes 
of arrays.  

=head2 EXPORT

None by default.
parse_table() is available if requested.

=head2 METHODS

Parse table is the only accessible method.  It is designed to accept either a filename or a
reference to an array of lines/strings as its first argument.  An optional second argument
may contain options such as the format of the data, its field separator, a request that its
return type of the method be a hash of arrays, and its comment symbol.

parse_table( $filename or \@data_lines,
           { format          => undef,  # File format (supported vals = fits|rdb|html|csv)
             field_separator => undef,  # Field separator for text file formats
             return_as_hash  => undef,  # Override default: return hash of arrays.
             return_as_pdl   => undef,  # Returns PDLs for numeric cols when return_as_hash=>1
             comment         => undef,  # Comment specifier (e.g. qr/^#/)
	     debug           => 0,      # print debug info if 1
           } 
);

In the absence of options specifying format and field separator, these
will be guessed by the method.  To guess the field separator, a
subroutine steps through these field separator choices in this order:
(",", "&", "|", "\t", qr/\s+/) .  If a separator separates all the
text lines into the same number of fields and that number is greater
than one, that separator is picked.  If no separator cleanly splits
into more than 1 column, the last separator that returns 1 column is
returned.  If no separator cleanly splits all of the lines to the same
number of columns, the routine will croak.


Absence of a comment specifier when used with a file with comments
will probably result in failure of the method.  (We can't guess
comment specifiers).

=head1 AUTHOR

Jean Connelly, E<lt>jeanconn@localdomain<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Smithsonian Astrophysical Observatory

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.0 or,
at your option, any later version of Perl 5 you may have available.


=cut

