#!/usr/bin/env perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--


#####################################################################
#
# Yaxx: Yet Another Xray Xtractor
#
# Copyright (C) 2006 by the Smithsonian Astrophysical Observatory
# 
# This code is released under the GNU General Public License.  You may
# find a copy at L<http://www.fsf.org/copyleft/gpl.html>.
#
# Creator: Tom Aldcroft
#
#####################################################################

use warnings;
use strict;

# Load various packages
use POSIX qw(strftime strtod);
use Data::Dumper;
use Cwd;
use English;
use File::Copy;
use File::Basename;
use File::Path;
use File::Glob qw(:glob);
use File::Find;
use File::Spec;
use IO::All;
use Config::General;
use Clone;
use Hash::Merge qw(merge);
use Getopt::Long;
use Carp;
use HTML::Table;
use List::MoreUtils qw(uniq);
use Text::ParseWords;
use Regexp::Common;

# 'Local' packages (not on CPAN)
use Ska::RDB qw(read_rdb);
use Ska::IO qw(read_ascii);
use Ska::HashTable;
use Ska::Convert;
use CFITSIO::Simple;
use Data::ParseTable qw(parse_table);
use GrabEnv;

# Set up global file and directory variables
my $program_dir = dirname( io($PROGRAM_NAME)->rel2abs );
my ($program_name) = fileparse($PROGRAM_NAME, qr{\.pl|\.bin});
my $config_file = "${program_name}.cfg";
my $analysis_dir = cwd;		# Directory where yaxx command called
my @summary;			# Data for source fitting summary
my @sources;			# YaxxSource objects for every source in sample

# Check for ciao (existence and version) and setenv PGPLOT_DIR
# check_for_ciao();

# Read in yaxx options.  First Yaxx cmd line options, then defaults, then user-spec'd config file
# In addition, if the -test flag is given, include config from 'yaxx.cfg.test' if available
my %par = get_cmd_line_options();	# Yaxx command line options
my ($cfg_files, %cfg) = get_config();

# Set Directory where yaxx results go.  
# Actually chdir there and use cwd to get resolved path in case of links
my $output_dir = eval { my $dir = io($cfg{output_dir})->chdir; cwd };
croak "ERROR - no output directory $cfg{output_dir}\n" if ($@ or not defined $output_dir);

# Load the appropriate Yaxx modules based on the data type (e.g. Chandra or XMM)
eval "use Yaxx::$cfg{mission}";
die $@ if ($@);

init_log_and_message();		# initialize log file and message subroutine

# Print the list of config files (now that the messages are set up)
message(1,"Reading config file $_\n") foreach (@{$cfg_files});
set_environment();		# Set various environment / program options
make_param_dir();		# Create temp location for parameter files (local to this process)

my @src_list = get_object_data($par{objlist} || $cfg{objlist});	# Read list of source objects for fitting
my @bkg_list = get_object_data($cfg{bkglist});	# Read list of background objects for extraction

# Let Yaxx package know about all the bkg sources
BKG_SOURCE: foreach my $bkg_data (@bkg_list) {
    my $obsid = $bkg_data->{obsid} or next BKG_SOURCE;
    my $input_dir = get_input_dir($cfg{input_dir}, $obsid) or next SOURCE;
    chdir $output_dir;

    # Make a new bkg Yaxx object for the background object (as for normal sources)
    next BKG_SOURCE unless my $bkg = Yaxx->new(\%cfg,
						     { input_dir => $input_dir,
						       output_dir => $output_dir,
						       analysis_dir  => $analysis_dir,
						       process  => 1,
						     },
						     $bkg_data,
						    );
    Yaxx->add_bkg_source($bkg);
}


# Make the global array of sources.  This includes all sources in sample
# (even if not processed this time) so that next_source and prev_source
# are correct.  Note that 'ignored' sources *are* skipped at this stage.

my $last_source;
foreach my $src_data (@src_list) {
    my $obsid = $src_data->{obsid};

    # Using the input_dir parameter, discover where the input data live.
    # input_dir could be relative to analysis dir, so start from there
    chdir $analysis_dir;
    my $input_dir;
    unless ($input_dir = get_input_dir($cfg{input_dir}, $obsid)) {
	message(1, "ERROR - could not get input directory for obsid $obsid based on '$cfg{input_dir}'\n");
	confess $Yaxx::FATAL_ERROR;
    }

    chdir $output_dir;

    for my $fakeid (0 .. $cfg{fake_instances}) {
    # and use these values to initialize Yaxx objects.
	my $source = Yaxx->new(\%cfg,
			       { input_dir     => $input_dir,
				 output_dir    => $output_dir,
				 analysis_dir  => $analysis_dir,
				 config_file   => $config_file,
				 program_dir   => $program_dir,
				 preclean      => $par{preclean},
				 clean         => $par{clean},
				 fakeid        => $fakeid,
			       },
			       $src_data,
			      );

	# Skip source if there is an IGNORE file present for this obsid
	next if (-e $source->file('ignore_obs')
		 or -e $source->file('ignore_src'));

	# Set process flag based on whether source passes input filter criteria (obsid,srcid)
	$source->process( filter_src($src_data, \%par) );

	# Set the next_source and prev_source attributes
	push @sources, $source;
	if (defined $last_source) {
	    $last_source->next_source($source);
	    $source->prev_source($last_source);
	}
	$last_source = $source;
    }
}
# Now do the actual processing of each source

SOURCE: foreach my $source (@sources) {
    chdir $output_dir;

    # Skip source if process flag not set (could be in src_list data file)
    next SOURCE unless $source->process;

    # Make fake dataset (i.e. set of fake src directories) for each source
    if ($par{make_fake}) {
	$source->make_fakes($par{make_fake});
	next SOURCE;
    }

    next SOURCE if (defined $par{src} and $par{src} != $source->srcid);

    # If doing a summary, just extract relevant information and go to the next Obsid
    #   THIS SHOULD go in Yaxx just as for validate (in fact real summary work is
    #   already there)
    if ($par{summary}) {
	push @summary, get_summary($source);
	next SOURCE;
    }

    # Now the real work for each source to be processed

    chdir $output_dir;
    next if (-e $source->file('ignore_src')); # Skip if IGNORE file is present in src dir

    message(1, "\n");
    message(1, '*' x 70 . "\n");
    message(1, sprintf("***  %-60s  ***\n", $source->unique_identifier));
    message(1, '*' x 70 . "\n");
    message(1, '', { time=>1 });

    next if $par{dryrun};

    # Go through the "pipeline" of processing steps

  PROCESS_STEP: while (my $process_step = $source->next_process_step) {
	my @detector = $process_step->{detector} ? listify($process_step->{detector}) : ('');
	foreach my $detector (@detector) {
	    $source->detector($detector);
	    my $ok = $source->run_task( $process_step );

	    # Stopped_at could be OK, for instance if source is locked by another process
	    # or quick_check decides that it has been processed successfully
	    $source->stopped_at($process_step->{name}) unless ($ok or defined $source->stopped_at);

	    # However, unless the success_on_not_OK param is set, then a stop => failure
	    $source->failed_at($process_step->{name})  unless ($ok or defined $source->failed_at
							       or $process_step->{success_on_not_OK});
	}
    }

    message(1,(defined $source->failed_at ? "*** FAILURE" : "*** SUCCESS") .
	    " for " . $source->unique_identifier . "\n",
	    { time=>1 });
    $source->close_log();
}

if ($par{summary}) {
    make_summary_table(\@summary);
} else {
    make_html_report_index(\@sources);
}
    

exit(0);

##****************************************************************************
sub clean_up {
#  This gets called either by END or signal trap
##****************************************************************************
    my $sig = shift;

    # Force dmcoords process in Yaxx to die.  This way of doing it is
    # ugly.  Need to move the DMcoord_open routine into yaxx.pl and provide
    # dmcoord access to $source in the main loop by passing DMcoord object ref
    # to Yaxx object.  Then yaxx.pl can properly start and stop the
    # dmcoord instances.  Do that next release!
    {
	no warnings;
	undef $Yaxx::Chandra::DMcoord;
    }

    make_param_dir('clean');	# Clean the temp (local) CIAO param dir
    if ($sig) {
	print STDERR "Stopped because of signal $sig\n" ;
	exit(1);
    }
    return;
}
    
##****************************************************************************
sub get_summary {
# 
# Call Yaxx functions to generate summary of Sherpa fit data for
# each specified source 
##****************************************************************************
    my $source = shift;
    my @summary;

    my $summ = $source->summary($par{summary});
    if ($summ) {
	message(1, "Generated summary for " . $source->unique_identifier . "\n");
	push @summary, $summ;
    }

    return @summary;
}

##****************************************************************************
sub get_input_dir {
##****************************************************************************
    my ($format, $obsid) = @_;
    my $input_dir = sprintf($format, $obsid);
    if ($input_dir =~ /[*?]/) {  # input_dir is a glob.  Try to resolve it
	my @dirs = glob($input_dir);
	if (@dirs != 1) {
	    message(1, "ERROR - ". scalar @dirs ." input data directories match $input_dir\n");
	    return;
	}
	$input_dir = $dirs[0];
    }
    return File::Spec->rel2abs($input_dir);
}

##****************************************************************************
sub get_config {
# Get the system defaults, project, thread, and test config files
##****************************************************************************
    Hash::Merge::set_behavior( 'RIGHT_PRECEDENT' );
    my $config = {};
    my @cfg_files;

    # Specify info about system and project config files
    my @cfg_files1 = ( { path     => $program_dir,
			 type     => 'system',
		       },
		       { path     => $analysis_dir,
			 type     => 'project',
		       },		       
		     );

    foreach (@cfg_files1) {
	my ($more_cfg, $cfg_file) = read_config_file($_);
	$config = merge($config, $more_cfg);
	push @cfg_files, $cfg_file;
    }

    # Specify info about thread and test config files. 
    my $thread = $config->{thread} or die "Need a thread value in the system or project config file\n";
    my @cfg_files2 = ( { path    => [ "$analysis_dir/resources/$thread",
				      "$program_dir/resources/$thread" ],
			 type    => 'thread',
		       },
		       { path     => $analysis_dir,
			 required => 0,
			 name     => "${config_file}.test",
			 type     => 'test',
		       }
		     );


    foreach (@cfg_files2) {
	my ($more_cfg, $cfg_file) = read_config_file($_);
	$config = merge($config, $more_cfg);
	push @cfg_files, $cfg_file;
    }

    return \@cfg_files, %{ $config };
}

##****************************************************************************
sub read_config_file {
##****************************************************************************
    my %opt = ( name     => $config_file,
		required => 1,
		type     => '',
		%{shift()},
	      );

    my @path = listify($opt{path});
    my $cfg_file_read;
    my %cfg;
    my $file;
    
    for my $path (@path) {
	next unless -r ($file = "$path/$opt{name}");
	%cfg = ParseConfig(-ConfigFile => $file,
			   -IncludeRelative => 1
			  );
	$cfg_file_read = 1;
	last;
    }

    die "Unable to find required $opt{type} config file $opt{name} in ", join(":",@path), "\n"
      unless ($cfg_file_read or not $opt{required});

    return \%cfg, $file;
}

##****************************************************************************
sub get_cmd_line_options {
##****************************************************************************
    my %par;

    my $status = GetOptions(\%par,
			    "obsid=s",
			    "src=i",
			    "preclean=s",
			    "summary=s",
			    "objlist=s",
			    "test!",
			   );

    die "ERROR - Invalid command line option\n" unless ($status);

    return %par;
}

##****************************************************************************
sub check_for_ciao {
##****************************************************************************
    local $_;
    foreach (qw (ASCDS_INSTALL ASCDS_BIN ASCDS_LIB CALDB PGPLOT_DIR)) {
	unless (defined $ENV{$_}) {
	    message(0, "\nERROR - CIAO variable '$_' not defined.  Are you in the CIAO environment?\n");
	    die "\n";
	}
    }
}

##****************************************************************************
sub set_environment {
##****************************************************************************
    $| = 1;			# Set stdout flushing
    $SIG{INT} = \&clean_up;		# Make sur to clean up if ctrl-c is pressed
    $SIG{QUIT} = \&clean_up;		# Make sur to clean up if die'ing

   # Set up env. vars based on tool environment definitions (e.g. CIAO, SAS, ftools etc)
    foreach my $env_name (listify( $cfg{use_environment} )) {
	message(1, "Initializing $env_name environment\n");
	my $env = $cfg{environment}{$env_name};
	unless ($env) {
	    message(0, "ERROR - 'environment' system config parameter for '$env_name' not found\n");
	    die "\n";
	}
	foreach my $setenv (listify( $env->{setenv} )) {
	    %ENV = GrabEnv::grabenv('csh', "setenv $setenv");
	}
	foreach my $script (listify( $env->{script} )) {
	    eval { %ENV = GrabEnv::grabenv($env->{shell}, $script) };
	    if ($@) {
		message(0, "ERROR - failed to initialize $env_name environment using '$script'\n");
		die "\n";
	    }
	    message(5, "Successfully inited $env_name using $script\n");
	}
    }

    $ENV{SLANG_SCRIPT_PATH} = join ":", "$program_dir/slang", $ENV{SLANG_SCRIPT_PATH};
}      

##****************************************************************************
sub init_log_and_message {
##****************************************************************************
    my $dir = io($output_dir)->chdir;

    # Make a unique log file name, and create log directory if needed
    if ($cfg{log_file} and $cfg{log_file} =~ /\/$/) {
	-d $cfg{log_file} or
	  mkdir $cfg{log_file} or die "Could not make log file dir $cfg{log_file}";
	$cfg{log_file} .= strftime("%Y-%m-%d_%H:%M", localtime);
    }

    # Create a message object to handle all program output
    my $msg = Ska::Message->new(stdout  => 1,
				file    => $cfg{log_file},
				auto_cr => 0,
			       );
    # Initialize the Yaxx::message routine
    Yaxx::message($msg, $cfg{verbose_stdout}, $cfg{verbose_master_log});
}


##****************************************************************************
sub get_obsids {
##****************************************************************************
    my @objlist = @_;
    my $file;

    # if $cfg{obsid} is numbers and whitespace, assume it's a list of obsids
    my @par_obsid = split ' ', $cfg{obsid} if ($cfg{obsid} and $cfg{obsid} =~ /^[\s\d]+$/);

    # Extract the obsids from the object list
    my %objlist_obsid = map { $_->{obsid} => 1 } @objlist;

    # If obsids specified in param, then intersect with object list, otherwise
    # use all obsids in object list
    my @obsid = @par_obsid ? 
	grep { exists $objlist_obsid{$_} } @par_obsid : keys %objlist_obsid;

    message(1, "Processing following obsids: @obsid\n", %par);

    return @obsid;
}


##****************************************************************************
sub get_object_data {
##****************************************************************************
    my $file = shift;
    my %count;
    my @data_out;		# Output data

    my %yaxx_fields = ( 'x'   => 'X',
			'y'   => 'Y',
			ra    => 'ra',
			dec   => 'dec',
			src   => 'srcid',
			ccdid => 'ccdid',
			obsid => 'obsid',
			rad   => 'rad',
			component => 'srcid',
			object => 'object',
		      );

    local $_;

    # If the supplied file name is empty just return
    return unless (defined $file and $file =~ /\S/);

    croak "ERROR - can't find object_list file $file\n" unless (-r $file);

    message(1, "Reading object list from $file\n");
    # Read the object data in as an array of hashes
    my @data_in = @{ parse_table($file) }; # This reads FITS, RDB, HTML, or ASCII

    # Make sure we got some data
    unless (@data_in) {
	message(0, "ERROR - object list file '$file' is empty");
	confess $Yaxx::FATAL_ERROR;
    }	

    # There must be an 'obsid' column
    if (not defined $data_in[0]->{obsid}) {
	message(0, "ERROR - need an 'obsid' column in object list file '$file'\n");
	confess $Yaxx::FATAL_ERROR;
    }		  

    # If none of the columns ra,dec,x,y exist, then bring in src2 file data
    unless (grep /^(ra|dec|x|y)$/i, keys %{$data_in[0]}) {
	@data_in = include_src_file_data(@data_in); # Expand any refs to detect file data
    }

    # Process each object data entry, making sure there is at least an obsid and srcid
    foreach my $data_in (@data_in) {
	# The key/val pairs in $data_out go directly into Yaxx object, so only
	# certain allowed values are copied in.  But put all data are in the "source_data"
	# attribute for use in user-defined fit scripts etc
	my $data = { source_data => $data_in };

	# Copy certain data into Yaxx attributes as spec'd in %yaxx_fields.
	# For convenience make the input matching case-independent so user could
	# have "RA Dec" in their input file and it will still fly
	foreach (keys %{$data_in}) {
	    my $yaxx_key = $yaxx_fields{ lc $_ };
	    $data->{$yaxx_key} = $data_in->{$_} if (defined $yaxx_key);
	}

	# Look for obsid, ccdid, srcid and clean them up
	for (qw(obsid ccdid srcid)) {
	    $data->{$_} = sprintf($cfg{formatting_rule}{$_}{fmt}, $data->{$_})
	      if defined $data->{$_};
	}
	
	# Generate a srcid if not provided
	$data->{srcid} = ++$count{ $data->{obsid} } unless defined $data->{srcid};

	push @data_out, $data;
    }

    return @data_out;
}

##****************************************************************************
sub include_src_file_data {
#
# Grab object data from file specified in the 'src_file' column of the objlist 
# input file.  This data will typically be a detect (wavdetect or celldetect)
# FITS file.  
#
# Output: array of merged object and source (detect) data
#
##****************************************************************************
    my @obj_data = @_;		# Object data from "objlist" file (array of hash-refs)
    my @obj_data_out;
    my %src_table_for;		# Hash of src (detect) file data based on obsid
    local $_;
    my @files;

    foreach my $obj_data (@obj_data) {
	my $obsid = $obj_data->{obsid};
	if (not defined $src_table_for{$obsid}) {
	    my $input_dir = get_input_dir($cfg{input_dir}, $obsid);

	    foreach (split ' ',$cfg{src2_glob}) {
		last if (@files = glob($input_dir."/$_") and not $!);
	    }

	    # Make sure we got one and only one event file
	    if ($! or @files > 1 or @files == 0) {
		my $err = (@files > 1) ? 'too many' : 'no';
		message(0, "ERROR - $err src2 files matching '".$cfg{src2_glob}."' for obsid $obsid\n");
		confess $Yaxx::FATAL_ERROR;
	    } 

	    $src_table_for{$obsid} = Ska::HashTable->new($files[0]); ;
	}

	# Combine $obj_data (obj_list file) and $data (from FITS
	# source file), with $obj_data taking precedence.  Lowercase all keys
	# to prevent possible errors from capitalized col names (e.g. RA). 
	my $src_col = $cfg{src_column_name};
	while (my %data = $src_table_for{$obsid}->row()) {
	    if (not defined $obj_data->{src}
		or $obj_data->{src} !~ /^\d+$/
		or $obj_data->{src} == $data{$src_col}) {

		# Skip source if not enough net counts
		next if (defined $data{net_counts} and $data{net_counts} < $cfg{min_counts});

		# Start with values from src (detect) file
		my %obj_src_data = map { lc($_) => $data{$_} } keys %data;

		# Override with values from object (objlist) file, but only if the new
		# value is non-trivial or if the existing value (from src) is undefined
		foreach (keys %{$obj_data}) {
		    my $lc_key = lc($_);
		    if ($obj_data->{$_} =~ /\S/ or not defined $obj_src_data{$lc_key}) {
			$obj_src_data{$lc_key} = $obj_data->{$_};
		    }
		}
		push @obj_data_out, { %obj_src_data };
	    }
	}
    }
    return @obj_data_out;
									   }

##****************************************************************************
sub make_html_report_index {
##****************************************************************************
    my $sources = shift;
    return unless @{$sources};	# Make sure there is something to do

    message(3, "\nCreating HTML report index\n");
    chdir $output_dir;

    # Not great, but use methods from first source for help
    my $source0 = $sources->[0]; 
    my @cols = listify($source0->html_report_index->{column});

    # Generated table header, using either formatting_rule label or capitalized var name
    my @table_data = ( [ map { $_->{header} } @cols ] );
  SOURCE:
    for my $source (@{$sources}) {

	# Get a copy of the stored (via Dumper) object in case current source
	# object was not processed this time
	unless ($source->process) {
	    my $new_source;
	    unless ($new_source = $source->restore_object) {
		message(1, "WARNING - could not restore source for $source->unique_identifier\n");
		next SOURCE;
	    }
	    $source = $new_source;
	}

	my @row_data;
	for my $col (@cols) {
	    # Put value into index table.  They might not be defined if processing failed,
	    # so suppress the 'use of uninitialized var' warnings
	    no warnings;
	    my $col_value = $source->interpolate_string($col->{value});
	    push @row_data, q{<a href="} . $source->file('report.html') . qq{"> $col_value </a>};
	}
	push @table_data, \@row_data;
    }

    my $html_table = HTML::Table->new(-class => 'style1',
				      -data => \@table_data);
    $html_table->setRowHead(1);

    my $report_index_html < io($source0->file('report_index_template.html'));
    eval {
	$report_index_html = $source0->interpolate_string($report_index_html,
							  INDEX_TABLE => $html_table->getTable,
							 );
    };
    message(2, "ERROR - $@\n") if $@;

    # Write the new index HTML content to the index file
    $report_index_html > io( $source0->file('report_index.html') );
}

##****************************************************************************
sub make_summary_table {
##****************************************************************************
    my $summ = shift;
    my $s;
    my $c;
    my @col;		# Ordered list of columns from summary->{cols}
    my %data;
    my $NAN = (-1)**0.5;	# Generate a NaN for undefined elements in summary table

    # Find all the columns.  {data} hash may have additional upper and lower limit cols
    @col = uniq map { @{$_->{cols}} } @{$summ};

    # collect the data into the structures for writing to FITS file
    foreach $c (@col) {
	foreach $s (@{$summ}) {
	    push @{$data{$c}}, defined ($s->{data}->{$c}) ? $s->{data}->{$c} : $NAN;
	}
    }

    if (fits_write_bintbl("$analysis_dir/summ_$par{summary}.fits", %data, {col_names => \@col})) {
	message(3, "Wrote summary for model $par{summary} to summ_$par{summary}.fits\n");
    } else {
	message(2, "ERROR - failed to write summary for model $par{summary} to summ_$par{summary}.fits\n");
    }
}

##****************************************************************************
sub filter_src {
##****************************************************************************
    my $src_data = shift;
    my $par = shift;
    local $_;

    # If input parameter set to a non-trivial value, take it as a comma-separated
    # list of allowed values.  Make corresponding sure $src_data vals match
    for (qw(obsid srcid)) {
	if ($par->{$_}) {
	    my %filter = map { $_ => 1 } split(/\s*,\s*/, $par->{$_});
	    return 0 unless $filter{ $src_data->{$_} };
	}
    }
    return 1;
}

##****************************************************************************
sub make_param_dir {
##****************************************************************************
    my $clean = shift;
    my $dir = "${analysis_dir}/cxcds_param_$PROCESS_ID";

    if ($clean) {
	return unless (-e $dir);
	eval { io($dir)->rmtree() };
	message("Failed to clean temporary param dir $dir: $@\n") if $@;
    } else {
	die "PFILES env variable is empty" unless $ENV{PFILES};
	my @pfiles = split ';', $ENV{PFILES};
	mkdir $dir or die "Could not make parameter file dir $dir: $!";
	$ENV{PFILES} = join ';', ($dir, $pfiles[-1]);
    }
}

########################################################################################
sub listify {
########################################################################################
    if (@_ and defined $_[0]) {
	return ref($_[0]) eq 'ARRAY' ? @{$_[0]} : ($_[0]);
    } else {
	return ();
    }
}

########################################################################################
sub message {
########################################################################################
    Yaxx::message(@_);
}

END {
    clean_up();
}
